package org.greenstone.atea;


import java.io.*;
import java.util.Properties;
import java.util.zip.GZIPInputStream;
import java.util.Iterator;
import java.util.Set;
import java.util.TreeSet;

import org.apache.log4j.Logger;

/**
 * The main() method of this class takes a folder of warc.wet(.gz) files and goes through
 * the WET records in each, putting each WET record into a file. Each file is put into a
 * keep or discard folder, based on content-length and number of lines.
 * A single instance of the WETProcessor class processes a single unzipped warc.wet file.
 *
 * To compile, including the jars in lib/ for compiling.
 *      maori-lang-detection/src$ javac -cp ".:../lib/*" org/greenstone/atea/CCWETProcessor.java 
 *
 * To run, passing the log4j and other properties files in conf/ folder:
 *      maori-lang-detection/src$ java -cp ".:../conf:../lib/*" org.greenstone.atea.CCWETProcessor &lt;folder containing warc.wet(.gz) files&gt; &lt;outputFolder&gt;
 *
 * e.g.
 *    - java -cp ".:../conf:../lib/*" org.greenstone.atea.CCWETProcessor ../tmp/processWET /Scratch/ak19/gs3-extensions/maori-lang-detection/tmp/processedWET
 *    - java -cp ".:../conf:../lib/*" org.greenstone.atea.CCWETProcessor ../tmp/processWET /Scratch/ak19/gs3-extensions/maori-lang-detection/tmp/processedWET 2&gt;&amp;1 | less
 *
*/

public class CCWETProcessor {
    private static Logger logger = Logger.getLogger(org.greenstone.atea.CCWETProcessor.class.getName());

    // Properties shared across WETProcessor instances
    public final int MAX_WORD_LENGTH;
    public final int MIN_NUM_WORDS;
    public final int MAX_WORDS_CAMELCASE;
    
    private Properties configProperties = new Properties();

    // File paths shared across WETProcessor instances
    public final File WETFilesDir;
    public final File outputFolder;
    public final File discardFolder;
    public final File keepFolder;
    public final File keepURLsFile;
    public final File discardURLsFile;

    // Keep a count of all the records that all WETProcessors instantiated
    // by our main method combined have processed
    private int totalRecordCount = 0;

    private int wetFileCount = 0;
    
    public CCWETProcessor(File inFolder, File outFolder) {
	this.WETFilesDir = inFolder;
	this.outputFolder = outFolder;

	// load up the properties from the config file
	try (InputStream infile = org.greenstone.atea.WETProcessor.class.getClassLoader().getResourceAsStream("config.properties")) {
	    configProperties = new Properties();
	    configProperties.load(infile);	    
	    //infile.close();
	    
	} catch(Exception e) {
	    System.err.println("Exception attempting to read properties from config.properties.");
	    logger.error("Exception attempting to read properties from config.properties.");
	    e.printStackTrace();
	}

	if(configProperties.size() == 0) {
	    System.err.println("*** Warning: no values read into config properties. Using defaults.");
	}
	
	MAX_WORD_LENGTH = Integer.parseInt(configProperties.getProperty("WETprocessor.max.word.length", "15"));
	MIN_NUM_WORDS = Integer.parseInt(configProperties.getProperty("WETprocessor.min.num.words", "20"));
	MAX_WORDS_CAMELCASE = Integer.parseInt(configProperties.getProperty("WETprocessor.max.words.camelcase", "10"));


	this.discardFolder = new File(outFolder, "discard");
	if(!discardFolder.exists()) {
	    discardFolder.mkdir();
	}	
	this.keepFolder = new File(outFolder, "keep");
	if(!keepFolder.exists()) {
	    keepFolder.mkdir();
	}

	this.keepURLsFile = new File(outFolder, "keepURLs.txt");	
	if(keepURLsFile.exists() && !keepURLsFile.delete()) {
	    System.err.println("Warning: Unable to delete " + this.keepURLsFile + ". Unable to proceed.");
	    //return;
	}
	this.discardURLsFile = new File(outFolder, "discardURLs.txt");
	if(discardURLsFile.exists() && !discardURLsFile.delete()) {
	    System.err.println("Warning Unable to delete " + discardURLsFile + ". Unable to proceed.");
	    //return;
	}
	
    }
    
    /**
     * Takes as input the keepURLs.txt file generated by running WETProcessor instances.
     * As output produces the URL seed list and regex-urlfilter text files required by nutch,
     * https://cwiki.apache.org/confluence/display/nutch/NutchTutorial
     */
    public void createSeedURLsFiles(File seedURLsFile, File urlFilterFile) {
	// Maintain Sets of unique domains and urls
	// TreeSet: by default, "the elements are ordered using their natural ordering"
	// (or by a Comparator provided at set creation time).
	// Whereas HashSet doesn't guarantee ordering.
	// So we get alphabetic sorting for free. And guaranteed log(n) for basic operations.

	Set<String> domainsSet = new TreeSet<String>();
	Set<String> urlsSet = new TreeSet<String>();

	final String FILTER_REGEX_PREFIX = "+https?://([a-z0-9-]+\\.)*"; // https?://([a-z0-9-]+\.)*
	
	try (
	     BufferedReader reader = new BufferedReader(new FileReader(this.keepURLsFile));
	     ) {
		
	    // read a URL at a time from urlsFile
	    String url = null;
	    String domain = null;
	    while((url = reader.readLine()) != null) { // readLine removes newline separator
		
		// work out domain. This retains any www. or subdomain prefix:
		int startIndex = url.indexOf("//"); // http:// or https:// prefix
		startIndex = (startIndex == -1) ? 0 : (startIndex+2); // skip past the protocol's // portion
		domain = url.substring(startIndex);
		int endIndex = domain.indexOf("/");
		if(endIndex == -1) endIndex = domain.length();
		domain = domain.substring(0, endIndex);

		//if(!domainsMap.containsKey(domain)) {
		urlsSet.add(url);
		domainsSet.add(domain);
		//}
	    }
	} catch (IOException ioe) {
	    ioe.printStackTrace();
	    System.err.println("\n@@@@@@@@@ Error reading in urls from file " + this.keepURLsFile);
	}
	
	try (BufferedWriter seedURLsWriter = new BufferedWriter(new FileWriter(seedURLsFile))) {
	    Iterator<String> i = urlsSet.iterator();
	    while(i.hasNext()) {
		String url = i.next();
		seedURLsWriter.write(url + "\n");
	    }
	    
	} catch (IOException ioe) {
	    ioe.printStackTrace();
	    System.err.println("\n@@@@@@@@@ Error writing to " + seedURLsFile);
	}

	try (BufferedWriter urlFilterWriter = new BufferedWriter(new FileWriter(urlFilterFile))) {
	    Iterator<String> i = domainsSet.iterator();
	    // nutch.apache.org => +^https?://([a-z0-9-]+\.)*nutch\.apache\.org/
	    while(i.hasNext()) {
		String domain = i.next();
		domain = FILTER_REGEX_PREFIX + domain.replace(".", "\\.") + "/";		
		urlFilterWriter.write(domain + "\n");
	    }
	    
	} catch (IOException ioe) {
	    ioe.printStackTrace();
	    System.err.println("\n@@@@@@@@@ Error writing to " + urlFilterFile);
	}
    }

    /*
    public boolean isBlacklisted(String url) {
	return false;
    }
    */

    public boolean isGreylisted(String url) {
	// alexa top sites and auto-translated product sites
	return false;
    }
    
    /**
     * Checks URL parameter against each line ("filter") of conf/url-discard-filter.txt to decide
     * whether it is in the discard list.
     * Filters don't represent actual regex, just ^ and $ as start and end terminators.
     * By not having this method deal with actual regex for filters, this has the advantage that
     * we don't have to remember to escape or double escape each filter to turn it into a regex.
     */
    //public boolean isInDiscardFilter(String url) {

    public boolean isBlacklisted(String url) {	
	String discardFilterFile = "url-discard-filter.txt"; // in conf folder

	try (
	     BufferedReader reader = new BufferedReader(new InputStreamReader(org.greenstone.atea.WETProcessor.class.getClassLoader().getResourceAsStream(discardFilterFile), "UTF-8"));
	     ) {
	    String filter = null;
	    while((filter = reader.readLine()) != null) {
		if(filter.trim().equals("")) {
		    continue;
		}
		//System.err.println("Got filter: " + filter);
		if(filter.startsWith("^") && filter.endsWith("$") && url.equals(filter.substring(1, filter.length()-1))) {
		    System.err.println("*** Discarding url " + url + "\n\tas it MATCHES filter " + filter);
		}
		else if(filter.startsWith("^") && url.startsWith(filter.substring(1))) {
		    System.err.println("*** Discarding url " + url + "\n\tas it STARTS WITH filter " + filter);
		    return true;
		}
		else if(filter.endsWith("$") && url.endsWith(filter.substring(0, filter.length()-1))) {
		    System.err.println("*** Discarding url " + url + "\n\tas it ENDS WITH filter " + filter);
		    return true;
		}
		else if(url.contains(filter)) {
		    System.err.println("*** Discarding url " + url + "\n\tas it CONTAINS filter " + filter);
		    return true;
		}
				
	    }
		
	} catch (IOException ioe) {
	    ioe.printStackTrace();
	    System.err.println("\n@@@@@@@@@ Error reading from " + discardFilterFile);
	}

	return false;
    }

    /** Maintain a count of all WET files processed. */
    public void setWETFileCount(int count) { this.wetFileCount = count; }
    
    /** Maintain a count of all WET records processed. */
    //public int getRecordCount() { return this.totalRecordCount; }
    //public void addToRecordCount(int count) { this.totalRecordCount += count; }
    public void setRecordCount(int count) { this.totalRecordCount = count; }
    
    public static void printUsage() {
	System.err.println("Run this program as:");
	System.err.println("\tWetProcessor <folder containing wet(.gz) files> <output folder path>");	
    }

    /** Filename filter to only list warc.wet files or else warc.wet.gz files
     * for which unzipped warc.wet equivalents don't yet exist.
     */
    private static class WETFilenameFilter implements FilenameFilter {
	
	public boolean accept(File dir, String name) {
	    if(name.endsWith(".warc.wet")) {
		logger.debug("Will include " + name + " for processing.");
		return true;
	    }

	    if(name.endsWith(".warc.wet.gz")) {
		String nameWithoutGZext = name.substring(0, name.lastIndexOf(".gz"));
		File unzippedVersion = new File(dir, nameWithoutGZext);
		if(unzippedVersion.exists()) {
		    logger.debug("--- Unzipped version " + unzippedVersion + " exists.");
		    logger.debug("Skipping " + name);
		    return false; // don't count gzipped version if unzipped version exists.
		}
		else {
		    logger.debug("Only zipped version " + name + " exists.");
		    return true; // No unzipped version, so have to work with gzipped version
		}
	    }

	    // we're not even interested in any other file extensions
	    logger.debug("Not a WET file. Skipping " + name);
	    return false;
	}
    }
	

    public static void main(String[] args) {
	if(args.length != 2) {
	    printUsage();
	    return;
	}
	
	
	File WETFileDir = new File(args[0]);
	if(!WETFileDir.exists() || !WETFileDir.isDirectory()) {
	    System.out.println("Error: " + args[0] + " does not exist or is not a directory");
	    return;
	}
	
	File outFolder = new File(args[1]);
	if(!outFolder.exists() || !outFolder.isDirectory()) {
	    System.out.println("Error: " + args[1] + " does not exist or is not a directory.");
	    return;
	}	

	CCWETProcessor ccWETFilesProcessor = new CCWETProcessor(WETFileDir, outFolder);

	//ccWETFilesProcessor.processAllWETFiles();
	
	// Will list all the warc.wet files in the input directory or else their gzipped versions
	File[] WETFiles = WETFileDir.listFiles(new WETFilenameFilter());

	int wetRecordCount = 0;
	int wetFileCount = 0;
	
	for(int i = 0; i < WETFiles.length; i++) {
	    File WETFile = WETFiles[i];	    
	    logger.debug("Processing WETfile: " + WETFile);

	    // Any .gz files listed means they haven't been unzipped yet. So unzip.
	    String WETFilename = WETFile.toString();
	    if(WETFilename.endsWith(".gz")) {
		File GZippedWETFile = WETFile;
		String WETGZippedFilename = WETFilename;
		WETFilename = WETFilename.substring(0, WETFilename.lastIndexOf(".gz"));		

		WETFile = new File(WETFilename);
		Utility.unzipFile(GZippedWETFile, WETFile);
	    }
	    // hereafter all WETFiles should refer to the unzipped version
	    // Check the unzipped WETFile exists	    

	    if(!WETFile.exists() || !WETFile.isFile()) {
		System.err.println("Error: " + WETFile + " does not exist (failure to unzip?)");
		logger.error("Error: " + WETFile + " does not exist (failure to unzip?)");
		return;
	    }

	    // Finally, we can process this WETFile's records into the keep and discard pile
	    wetFileCount++;
	    logger.debug("Off to process " + WETFile);
	    WETProcessor wetFileProcessor = new WETProcessor(WETFile, ccWETFilesProcessor);
	    wetFileProcessor.processWETFile();
	    wetRecordCount += wetFileProcessor.getRecordCount();
	}

	// for information purposes
	ccWETFilesProcessor.setWETFileCount(wetFileCount);
	ccWETFilesProcessor.setRecordCount(wetRecordCount);
	
	File seedURLsFile = new File(outFolder, "seedURLs.txt");
	File urlFilterFile = new File(outFolder, "regex-urlfilter.txt");
	ccWETFilesProcessor.createSeedURLsFiles(seedURLsFile, urlFilterFile);
	
	return;	
    
    }
}
